//
//
//	3ds files parser class
//
//											(C) 1997 JoOl

#ifndef _PARSER_3DS_H_
#define _PARSER_3DS_H_

#include "ParserBin.h"


class					Parser3ds : public ParserBin
{

public:

	// Main chunk
	enum
	{
		MAIN_3DS				= 0x4d4d,		// 3ds binary files
		MAIN_MLI				= 0x3daa,		// Material library files
		MAIN_PRJ				= 0x3dc2		// 3ds project file
	};

	// Sub chunks of MAIN3DS
	enum
	{
		EDIT3DS					= 0x3d3d,
		KEYF3DS					= 0xb000
	};

	// Sub chunks of EDIT3DS
	enum
	{
		EDIT_MATERIAL			= 0xafff,
		EDIT_CONFIG1			= 0x0100,
		EDIT_CONFIG2			= 0x3e3d,
		EDIT_VIEW_P1			= 0x7012,
		EDIT_VIEW_P2			= 0x7011,
		EDIT_VIEW_P3			= 0x7020,
		EDIT_VIEW1				= 0x7001,
		EDIT_BACKGR				= 0x1200,
		EDIT_AMBIENT			= 0x2100,
		EDIT_OBJECT				= 0x4000,

		EDIT_UNKNOWN01			= 0x1100,
		EDIT_UNKNOWN02			= 0x1201,
		EDIT_UNKNOWN03			= 0x1300,
		EDIT_UNKNOWN04			= 0x1400,
		EDIT_UNKNOWN05			= 0x1420,
		EDIT_UNKNOWN06			= 0x1450,
		EDIT_UNKNOWN07			= 0x1500,
		EDIT_UNKNOWN08			= 0x2200,
		EDIT_UNKNOWN09			= 0x2201,
		EDIT_UNKNOWN10			= 0x2210,
		EDIT_UNKNOWN11			= 0x2300,
		EDIT_UNKNOWN12			= 0x2302,
		EDIT_UNKNOWN13			= 0x3000
	};

	// Sub chunks of EDIT_MATERIAL
	enum
	{
		MAT_NAME				= 0xa000,
		MAT_AMBIENTCOLOR		= 0xa010,
		MAT_DIFFUSECOLOR		= 0xa020,
		MAT_SPECULARCOLOR		= 0xa030,
		MAT_SHININESS			= 0xa040,
		MAT_TRANSPARENCY		= 0xa050,
		MAT_TRANSFALLOFF		= 0xa052,
		MAT_REFLECTBLUR			= 0xa053,
		MAT_TEXTUREMAP			= 0xa200,
		MAT_OPACITYMAP			= 0xa210,
		MAT_TRANSFOFFENABLE		= 0xa240,
		MAT_REFLBLURENABLE		= 0xa250,
		MAT_BUMPMAP				= 0xa230,
		MAT_MAPFILE				= 0xa300
	};

	// Sub chunks of EDIT_OBJECT
	enum
	{
		OBJ_TRIMESH				= 0x4100,
		OBJ_LIGHT				= 0x4600,
		OBJ_CAMERA				= 0x4700,

		OBJ_UNKNOWN1			= 0x4010,
		OBJ_UNKNOWN2			= 0x4012			// shadow ?
	};

	// Sub chunks of OBJ_CAMERA
	enum
	{
		CAM_UNKNOWN1			= 0x4710,
		CAM_UNKNOWN2			= 0x4720
	};

	// Sub chunks of OBJ_LIGHT
	enum
	{
		LIGHT_OFF				= 0x4620,
		LIGHT_SPOT				= 0x4610,
		LIGHT_UNKNOWN			= 0x465a
	};

	// Sub chunks of OBJ_TRIMESH
	enum
	{
		TRI_VERTLIST			= 0x4110,
		TRI_FACELIST2			= 0x4111,
		TRI_FACELIST			= 0x4120,
		TRI_MATERIAL			= 0x4130,
		TRI_MAPCOORD			= 0x4140,
		TRI_SMOOTH				= 0x4150,
		TRI_LOCALAXIS			= 0x4160,
		TRI_VISIBLE				= 0x4165,
		TRI_MAPPINGSTD			= 0x4170
	};


	// Sub chunks of KEYF3DS
	enum
	{
		KEYF_UNKNOWN1			= 0xb009,
		KEYF_UNKNOWN2			= 0xb00a,
		KEYF_FRAMES				= 0xb008,
		KEYF_OBJDESCR			= 0xb002
	};

	// Color chunks
	enum
	{
		COLOR_RGBFLOAT			= 0x0010,
		COLOR_RGBBYTE			= 0x0011,
		COLOR_UNKNOWN			= 0x0013
	};

	// Viewport chunks
	enum
	{
		VIEWPORT_TOP			= 0x0001,
		VIEWPORT_BOTTOM			= 0x0002,
		VIEWPORT_LEFT			= 0x0003,
		VIEWPORT_RIGHT			= 0x0004,
		VIEWPORT_FRONT			= 0x0005,
		VIEWPORT_BACK			= 0x0006,
		VIEWPORT_USER			= 0x0007,
		VIEWPORT_CAMERA			= 0xffff,
		VIEWPORT_LIGHT			= 0x0009
	};

private:

	class				ChunkList
	{
		class			Chunk
		{
			int32				Id;
			const char*			Lbl;

		public:
			Chunk(int32 _id, const char* _lbl)
			{
				Id = _id;
				Lbl = _lbl;
			};
			~Chunk()		{};

			int32			GetId()
			{ return Id;	};
			const char*		GetName()
			{ return Lbl;	};
		};


		int32			SearchedId;
		BList			List;

	public:

		ChunkList()
		{
			List.AddItem(new Chunk(MAIN_3DS, "MAIN_3DS"));
				List.AddItem(new Chunk(EDIT3DS, "EDIT3DS"));
					List.AddItem(new Chunk(EDIT_MATERIAL, "EDIT_MATERIAL"));
						List.AddItem(new Chunk(MAT_NAME, "MAT_NAME"));
						List.AddItem(new Chunk(MAT_AMBIENTCOLOR, "MAT_AMBIENTCOLOR"));
						List.AddItem(new Chunk(MAT_DIFFUSECOLOR, "MAT_DIFFUSECOLOR"));
						List.AddItem(new Chunk(MAT_SPECULARCOLOR, "MAT_SPECULARCOLOR"));
						List.AddItem(new Chunk(MAT_SHININESS, "MAT_SHININESS"));
						List.AddItem(new Chunk(MAT_TRANSPARENCY, "MAT_TRANSPARENCY"));
						List.AddItem(new Chunk(MAT_TRANSFALLOFF, "MAT_TRANSFALLOFF"));
						List.AddItem(new Chunk(MAT_REFLECTBLUR, "MAT_REFLECTBLUR"));
						List.AddItem(new Chunk(MAT_TEXTUREMAP, "MAT_TEXTUREMAP"));
						List.AddItem(new Chunk(MAT_OPACITYMAP, "MAT_OPACITYMAP"));
						List.AddItem(new Chunk(MAT_TRANSFOFFENABLE, "MAT_TRANSFOFFENABLE"));
						List.AddItem(new Chunk(MAT_REFLBLURENABLE, "MAT_REFLBLURENABLE"));
						List.AddItem(new Chunk(MAT_BUMPMAP, "MAT_BUMPMAP"));
						List.AddItem(new Chunk(MAT_MAPFILE, "MAT_MAPFILE"));
					List.AddItem(new Chunk(EDIT_CONFIG1, "EDIT_CONFIG1"));
					List.AddItem(new Chunk(EDIT_CONFIG2, "EDIT_CONFIG2"));
					List.AddItem(new Chunk(EDIT_VIEW_P1, "EDIT_VIEW_P1"));
					List.AddItem(new Chunk(EDIT_VIEW_P2, "EDIT_VIEW_P2"));
					List.AddItem(new Chunk(EDIT_VIEW_P3, "EDIT_VIEW_P3"));
					List.AddItem(new Chunk(EDIT_VIEW1, "EDIT_VIEW1"));
					List.AddItem(new Chunk(EDIT_BACKGR, "EDIT_BACKGR"));
					List.AddItem(new Chunk(EDIT_AMBIENT, "EDIT_AMBIENT"));
					List.AddItem(new Chunk(EDIT_OBJECT, "EDIT_OBJECT"));
						List.AddItem(new Chunk(OBJ_TRIMESH, "OBJ_TRIMESH"));
							List.AddItem(new Chunk(TRI_VERTLIST, "TRI_VERTLIST"));
							List.AddItem(new Chunk(TRI_FACELIST2, "TRI_FACELIST2"));
							List.AddItem(new Chunk(TRI_FACELIST, "TRI_FACELIST"));
							List.AddItem(new Chunk(TRI_MATERIAL, "TRI_MATERIAL"));
							List.AddItem(new Chunk(TRI_MAPCOORD, "TRI_MAPCOORD"));
							List.AddItem(new Chunk(TRI_SMOOTH, "TRI_SMOOTH"));
							List.AddItem(new Chunk(TRI_LOCALAXIS, "TRI_LOCALAXIS"));
							List.AddItem(new Chunk(TRI_VISIBLE, "TRI_VISIBLE"));
							List.AddItem(new Chunk(TRI_MAPPINGSTD, "TRI_MAPPINGSTD"));
						List.AddItem(new Chunk(OBJ_LIGHT, "OBJ_LIGHT"));
							List.AddItem(new Chunk(LIGHT_OFF, "LIGHT_OFF"));
							List.AddItem(new Chunk(LIGHT_SPOT, "LIGHT_SPOT"));
							List.AddItem(new Chunk(LIGHT_UNKNOWN, "LIGHT_UNKNOWN"));
						List.AddItem(new Chunk(OBJ_CAMERA, "OBJ_CAMERA"));
							List.AddItem(new Chunk(CAM_UNKNOWN1, "CAM_UNKNOWN1"));
							List.AddItem(new Chunk(CAM_UNKNOWN2, "CAM_UNKNOWN2"));
						List.AddItem(new Chunk(OBJ_UNKNOWN1, "OBJ_UNKNOWN1"));
						List.AddItem(new Chunk(OBJ_UNKNOWN2, "OBJ_UNKNOWN2"));
					List.AddItem(new Chunk(EDIT_UNKNOWN01, "EDIT_UNKNOWN01"));
					List.AddItem(new Chunk(EDIT_UNKNOWN02, "EDIT_UNKNOWN02"));
					List.AddItem(new Chunk(EDIT_UNKNOWN03, "EDIT_UNKNOWN03"));
					List.AddItem(new Chunk(EDIT_UNKNOWN04, "EDIT_UNKNOWN04"));
					List.AddItem(new Chunk(EDIT_UNKNOWN05, "EDIT_UNKNOWN05"));
					List.AddItem(new Chunk(EDIT_UNKNOWN06, "EDIT_UNKNOWN06"));
					List.AddItem(new Chunk(EDIT_UNKNOWN07, "EDIT_UNKNOWN07"));
					List.AddItem(new Chunk(EDIT_UNKNOWN08, "EDIT_UNKNOWN08"));
					List.AddItem(new Chunk(EDIT_UNKNOWN09, "EDIT_UNKNOWN09"));
					List.AddItem(new Chunk(EDIT_UNKNOWN10, "EDIT_UNKNOWN10"));
					List.AddItem(new Chunk(EDIT_UNKNOWN11, "EDIT_UNKNOWN11"));
					List.AddItem(new Chunk(EDIT_UNKNOWN12, "EDIT_UNKNOWN12"));
					List.AddItem(new Chunk(EDIT_UNKNOWN13, "EDIT_UNKNOWN13"));
				List.AddItem(new Chunk(KEYF3DS, "KEYF3DS"));
					List.AddItem(new Chunk(KEYF_UNKNOWN1, "KEYF_UNKNOWN1"));
					List.AddItem(new Chunk(KEYF_UNKNOWN2, "KEYF_UNKNOWN2"));
					List.AddItem(new Chunk(KEYF_FRAMES, "KEYF_FRAMES"));
					List.AddItem(new Chunk(KEYF_OBJDESCR, "KEYF_OBJDESCR"));
			List.AddItem(new Chunk(COLOR_RGBFLOAT, "COLOR_RGBFLOAT"));
			List.AddItem(new Chunk(COLOR_RGBBYTE, "COLOR_RGBBYTE"));
			List.AddItem(new Chunk(COLOR_UNKNOWN, "COLOR_UNKNOWN"));
			List.AddItem(new Chunk(VIEWPORT_TOP, "VIEWPORT_TOP"));
			List.AddItem(new Chunk(VIEWPORT_BOTTOM, "VIEWPORT_BOTTOM"));
			List.AddItem(new Chunk(VIEWPORT_LEFT, "VIEWPORT_LEFT"));
			List.AddItem(new Chunk(VIEWPORT_RIGHT, "VIEWPORT_RIGHT"));
			List.AddItem(new Chunk(VIEWPORT_FRONT, "VIEWPORT_FRONT"));
			List.AddItem(new Chunk(VIEWPORT_BACK, "VIEWPORT_BACK"));
			List.AddItem(new Chunk(VIEWPORT_USER, "VIEWPORT_USER"));
			List.AddItem(new Chunk(VIEWPORT_CAMERA, "VIEWPORT_CAMERA"));
			List.AddItem(new Chunk(VIEWPORT_LIGHT, "VIEWPORT_LIGHT"));

			List.AddItem(new Chunk(MAIN_MLI, "MAIN_MLI"));

			List.AddItem(new Chunk(MAIN_PRJ, "MAIN_PRJ"));

			List.SortItems(CompareChunks);
		};

		~ChunkList()
		{
			for (int32 i = List.CountItems()-1; i >= 0; i--)
				delete (Chunk* )List.RemoveItem(int32(0));
		};

		static int			CompareChunks(const void* Chunk1, const void* Chunk2)
		{
			return (*(Chunk** )Chunk1)->GetId() - (*(Chunk** )Chunk2)->GetId();
		};

		const char*			Search(int32 Id)
		{
			SearchedId = Id;
			return SearchRecur(0, List.CountItems()-1);
		};

		const char*			SearchRecur(int32 i, int32 j)
		{
			int32			k = (i+j) >> 1;
			int32			CurrentId = ((Chunk* )List.ItemAt(k))->GetId();

			if (i == j)
			{
				if (CurrentId == SearchedId)
					return ((Chunk* )List.ItemAt(k))->GetName();
				return NULL;
			}
			if (CurrentId < SearchedId)
				return SearchRecur(k+1, j);
			else
				return SearchRecur(i, k);
		};
	};

	ChunkList				Chunks;


// Nothing but text!
#define INFO(a)				PRINT3(a ": %04x, %s %d\n", ChunkId,		\
									Chunks.Search(ChunkId), ChunkLen)

#define READ_CHUNKID		uint16			ChunkId;	\
							READ_INT16(ChunkId)

#define READ_CHUNKLEN		uint32			ChunkLen;	\
							READ_INT32(ChunkLen)

#define READ_CHUNK			READ_CHUNKID;	\
							READ_CHUNKLEN

#define SKIP				SkipChunk(ChunkId, ChunkLen)

#define CALL(a)				Indent -= 2;				\
							ASSERT(a(ChunkLen));		\
							Indent += 2

#define CALL2(a, b)			Indent -= 2;				\
							ASSERT(a(ChunkLen, b));		\
							Indent += 2

#define CHECK(a)			if (CurrLen != Len)		\
							{						\
								printf("%s@@@@@ " a ": %d/%d\n", Indent, CurrLen, Len);	\
								return false;		\
							}						\
							printf("%s " a ": %d/%d\n", Indent, CurrLen, Len);	\
							return true


	bool					SkipChunk(uint16, uint32);

	// private chunks parsing methods
	bool					ReadMainChunk(uint32);
	bool					ReadKeyfChunk(uint32);
	bool					ReadEditChunk(uint32);
	bool					ReadEditMaterialChunk(uint32);
	bool					ReadViewChunk(uint32);
	bool					ReadBackgroundChunk(uint32);
	bool					ReadAmbientChunk(uint32);
	bool					ReadObjectChunk(uint32);
	bool					ReadViewportChunk(uint32);
	bool					ReadRgbFloatChunk(uint32);
	bool					ReadRgbByteChunk(uint32);
	bool					ReadTriMeshChunk(uint32);
	bool					ReadLightChunk(uint32);
	bool					ReadCameraChunk(uint32);
	bool					ReadVerticesChunk(uint32);
	bool					ReadFacesChunk(uint32);
	bool					ReadFaceMaterialChunk(uint32);
	bool					ReadFaceSmoothChunk(uint32, uint16);
	bool					ReadTransfoChunk(uint32);
	bool					ReadSpotChunk(uint32);

public:

	Parser3ds(bool = false, int32 = LEXER_BUFLEN, int32 = LEXER_LEXMLEN);
	~Parser3ds();

	virtual bool			SetInput(entry_ref& );

	virtual bool			CheckType();
	virtual bool			Parse();
};


#endif	/* _PARSER_3DS_H_ */



