//
//
//	DXF file parser class
//
//											(C) 1997 JoOl


#ifndef _DXF_PARSER_H_
#define _DXF_PARSER_H_


#include "ParserText.h"
#include "LexerDxf.h"


class						ParserDxf: public ParserText
{
#define VERSION_R10			"AC1006"
#define VERSION_R11			"AC1009"
#define VERSION_R12			"AC1009"

#define PLINE_CLOSED		0x01	// closed polyline
#define PLINE_CFIT_VERT		0x02	// curve-fit vertices have been added
#define PLINE_SFIT_VERT		0x04	// spline-fit vertices have been added
#define PLINE_3D			0x08	// 3d Polyline
#define PLINE_POLYMESH3D	0x10	// 3d polygon mesh

#define VERTEX_CFITTING		0x01	// Extra vertex created by curve-fitting
#define VERTEX_CFIT_TAN		0x02	// Curve-fit tangent defined for this vertex
#define VERTEX_SFITTING		0x08	// Spline vertex created by spline-fitting
#define VERTEX_SCPOINT		0x10	// Spline frame control point
#define VERTEX_PLINE3D		0x20	// 3D Polyline vertex
#define VERTEX_MESH3D		0x40	// 3D polygon mesh vertex
#define VERTEX_PFACE		0x80	// Polyface mesh vertex


#define MATCH_GROUP(a, b)	if (!Match(a, b))		return false

#define SKIP_GROUP			ASSERT(NextToken())


	// Peeking and matching routines
	bool					NextToken();
	bool					Match(int32);
	bool					Match(int32, uint32);
	bool					Peek(int32);
	bool					Peek(int32, uint32);

	// a few useful routines
	bool					ParseTo(int32);
	bool					ParseTo(int32, uint32);
	bool					MatchGroup(int32, int32& );
	bool					MatchGroup(int32, float& );
	bool					MatchGroupCast(int32, float& );
	bool					MatchGroup(int32, char*& );
	bool					MatchGroupCast(int32, char*& );
	bool					ParseVector(int32, float& , float& , float& );

	void					PrintGroup();

	// Header-related routines
	bool					ParseHeader();
	bool					ParseVersion();

	// Table-related routines
	bool					ParseTables();
	bool					ParseTable();
	bool					ParseLayers();
	bool					ParseLayer();
	bool					ParseView();
	bool					ParseAppId();

	// Entities-related routines
	bool					ParseEntities();
	bool					ParseLine();
	bool					ParsePoint();
	bool					ParseVertex();
	bool					ParseFace3d();
	bool					ParsePolyline();


	int32					GroupCode;

public:

	ParserDxf(bool, int32 = LEXER_BUFLEN, int32 = LEXER_LEXMLEN);
	~ParserDxf();

	virtual bool			CheckType();
	virtual bool			Parse();
};


#endif	/* _DXF_PARSER_H_ */

