//
//
//	3d2 files parser class
//
//											(C) 1997 JoOl


#include "ParserIob.h"


ParserIob::ParserIob(bool V, int32 BufLen, int32 LexmLen) : ParserBin(V)
{
	Lexr = new LexerBin(BufLen, LexmLen);

	QuickDraw[0] = "none";
	QuickDraw[1] = "bounding box";
	QuickDraw[2] = "quick edges";

	PartShape[PART_SHAPE_FACE] = "face";
	PartShape[PART_SHAPE_TETRAHEDRON] = "tetrahedron";
	PartShape[PART_SHAPE_PYRAMID] = "pyramid";
	PartShape[PART_SHAPE_OCTAHEDRON] = "octahedron";	
	PartShape[PART_SHAPE_CUBE] = "cube";
	PartShape[PART_SHAPE_BLOCK] = "block";
	PartShape[PART_SHAPE_DODECAHEDRON] = "dodecahedron";
	PartShape[PART_SHAPE_SPHERE] = "sphere";
	PartShape[PART_SHAPE_RANDOM] = "random";	
	PartShape[PART_SHAPE_USE_OBJECT_FILE] = "use object file";

	PartCenter[0] = "inscribed";
	PartCenter[1] = "circumbscribed";
	PartCenter[2] = "interpolated";
	PartCenter[3] = "barycentric";

	PartSize[0] = "small";
	PartSize[1] = "large";
	PartSize[2] = "random";
	PartSize[3] = "specify";

	PartAlign[0] = "to object";
	PartAlign[1] = "to face";
	PartAlign[2] = "random";
}


ParserIob::~ParserIob()
{
	delete Lexr;
}


bool
ParserIob::CheckType()
{
	uint32			Len;
	ASSERT(ParseHeader(Len));

	return true;
}


bool
ParserIob::SetInput(entry_ref& _e)
{
	return inherited::SetInput(_e, P_BIG_ENDIAN);
}


//---------------------------------------
// Main parsing routine
//---------------------------------------

bool
ParserIob::Parse()
{
	uint32			CurrLen = 0;
	uint32			Len;
	ASSERT2(ParseHeader(Len), "header");

	while (CurrLen < Len)
	{
		READ_CHUNK;

		if (ChunkId == OBJECT_ID)
		{
			ASSERT2(ParseObject(ChunkLen), "obj");
		}
		else
		{
			SKIP_CHUNK;
		}

		CurrLen += ChunkLen + 2*sizeof(ChunkId);
	}

printf("********************* %d/%d\n", CurrLen, Len);
	Terminate();
	printf("Done...\n");

	CHECK("parse");
}



//---------------------------------------
// Header parsing routine
//---------------------------------------

bool
ParserIob::ParseHeader(uint32& Out)
{
	off_t				FileSiz;
	int32				FormType;

	Lexr->GetInputSize(FileSiz);
	READ_CHUNK;
	Out = ChunkLen;
	READ_INT32(FormType);

	ASSERT(ChunkId == FORM_ID);
	ASSERT(ChunkLen + 2*sizeof(ChunkId) == int32(FileSiz));
	ASSERT(FormType == FORM_TYPE);

	PRINT5("header '%c%c%c%c' %d", ((ChunkId >> 24) & 0x00ff),	
		((ChunkId >> 16) & 0x00ff), ((ChunkId >> 8) & 0x00ff),
		(ChunkId & 0x00ff), ChunkLen);
	_PRINT4(" Type='%c%c%c%c'\n", ((FormType >> 24) & 0x00ff),	
		((FormType >> 16) & 0x00ff), ((FormType >> 8) & 0x00ff),
		(FormType & 0x00ff));

	return true;
}



bool
ParserIob::ParseObject(uint32 Len)
{
	uint32			CurrLen = 0;

	while (CurrLen < Len)
	{
		READ_CHUNK;
		PRINT_CHUNK;

		if (ChunkId == OBJ_DESC)
		{
			ASSERT2(ParseObjectDesc(ChunkLen), "obj desc");
		}
		else
		{
			SKIP_CHUNK;
		}

		CurrLen += ChunkLen + 2*sizeof(ChunkId);
	}

	CHECK("obj");
}


bool
ParserIob::ParseObjectDesc(uint32 Len)
{
	uint32			CurrLen = 0;

	while (CurrLen < Len)
	{
		READ_CHUNK;

/*		if (ChunkId == )
		{
			ASSERT();
		}
		else
*/		{
			SKIP_CHUNK;
		}

		CurrLen += ChunkLen + 2*sizeof(ChunkId);
	}

	CHECK("obj desc");
}




//---------------------------------------
// A few useful routines
//---------------------------------------

bool
ParserIob::SkipChunk(uint32 Len)
{
	ASSERT(SkipBytes(Len));
	return true;
}


bool
ParserIob::ParseColor(uchar& r, uchar& g, uchar& b)
{
	READ_INT8(r);
	READ_INT8(g);
	READ_INT8(b);

	return true;
}


bool
ParserIob::ParseColorPad(uchar& r, uchar& g, uchar& b)
{
	uchar			Pad;
	READ_INT8(Pad);
	ASSERT(Pad == 0);

	READ_INT8(r);
	READ_INT8(g);
	READ_INT8(b);

	return true;
}


bool
ParserIob::ParseFract(float& Out)
{
	int32			Val;
	READ_INT32(Val);
	Out = float(Val) / 65536.0;

	return true;
}


bool
ParserIob::ParseVector(float& x, float& y, float& z)
{
	ASSERT(ParseFract(x));
	ASSERT(ParseFract(y));
	ASSERT(ParseFract(z));

	return true;
}


bool
ParserIob::ParseMatrix()
{
	float			x, y, z;

	ASSERT(ParseVector(x, y, z));
	PRINT3("matrix: %f %f %f\n", x, y, z);
	ASSERT(ParseVector(x, y, z));
	PRINT3("  %f %f %f\n", x, y, z);
	ASSERT(ParseVector(x, y, z));
	PRINT3("  %f %f %f\n", x, y, z);

	return true;
}



bool
ParserIob::ParseTform(uint32 Len)
{
	ASSERT(Len == 60);

	float			x, y, z;
	ASSERT(ParseVector(x, y, z));
	PRINT3("position (%f %f %f)\n", x, y, z);
	ASSERT(ParseVector(x, y, z));
	PRINT3("x axis (%f %f %f)\n", x, y, z);
	ASSERT(ParseVector(x, y, z));
	PRINT3("y axis (%f %f %f)\n", x, y, z);
	ASSERT(ParseVector(x, y, z));
	PRINT3("z axis (%f %f %f)\n", x, y, z);
	ASSERT(ParseVector(x, y, z));
	PRINT3("size (%f %f %f)\n", x, y, z);

	return true;
}


//---------------------------------------
// Name-related routines
//---------------------------------------

bool
ParserIob::ParseName(uint32 Len, char* Out)
{
	ASSERT(ReadBytes(Len, Out));
	PRINT1("name '%s'\n", Out);

	return true;
}


//---------------------------------------
// Shp2-related routines
//---------------------------------------
/*
bool
ParserIob::ParseShp2(uint32 Len)
{
	ASSERT(Len == 4);
	int16			Shape;
	int16			Lamp;
	READ_INT16(Shape);
	PRINT0("shp2: shape=");		PrintShp2Shape(Shape);
	READ_INT16(Lamp);
	PRINT0(" / lamp=");		PrintShp2Lamp(Lamp);
	_PRINT0("\n");

	return true;
}


void
ParserIob::PrintShp2Shape(uint16 Shape)
{
	if (Shape == SHP2_TYPE_SPHERE)
	{
		_PRINT0("sphere");
	}
	else if (Shape == SHP2_TYPE_AXIS)
	{
		_PRINT0("axis");
	}
	else if (Shape == SHP2_TYPE_GROUND)
	{
		_PRINT0("ground");
	}
	else
	{
		_PRINT0("<unknown>");
	}
}


void
ParserIob::PrintShp2Lamp(uint16 Lamp)
{
	if (Lamp & SHP2_LAMP_POINT_SOURCE)
	{
		_PRINT0("point source, ");
	}
	else if (Lamp & SHP2_LAMP_PARALLEL_SOURCE)
	{
		_PRINT0("parralel source, ");
	}

	if (Lamp & SHP2_LAMP_ROUND_SHAPE)
	{
		_PRINT0("round shape, ");
	}
	else if (Lamp & SHP2_LAMP_RECT_SHAPE)
	{
		_PRINT0("rectangualar shape, ");
	}

	if (Lamp & SHP2_LAMP_LENS_FLARE)
	{
		_PRINT0("lens flare, ");
	}
	if (Lamp & SHP2_LAMP_DIMINISH_I)
	{
		_PRINT0("diminish intensity, ");
	}
	if (Lamp & SHP2_LAMP_FALLOFF)
	{
		_PRINT0("falloff, ");
	}
	if (Lamp & SHP2_LAMP_CAST_SHADOWS)
	{
		_PRINT0("cast shadows, ");
	}
	if (Lamp & SHP2_LAMP_BRIGHT_OBJ)
	{
		_PRINT0("bright object, ");
	}
}


bool
ParserIob::ParsePosi(uint32 Len)
{
	ASSERT(Len == 12);
	float			x, y, z;

	ASSERT(ParseVector(x, y, z));
	PRINT3("object at (%f %f %f)\n", x, y, z);

	return true;
}


bool
ParserIob::ParseAxis(uint32 Len)
{
	ASSERT(Len == 36);
	float			x, y, z;

	ASSERT(ParseVector(x, y, z));
	PRINT3("object u (%f %f %f)\n", x, y, z);
	ASSERT(ParseVector(x, y, z));
	PRINT3("object v (%f %f %f)\n", x, y, z);
	ASSERT(ParseVector(x, y, z));
	PRINT3("object n (%f %f %f)\n", x, y, z);

	return true;
}


bool
ParserIob::ParseSize(uint32 Len)
{
	ASSERT(Len == 12);
	float			x, y, z;

	ASSERT(ParseVector(x, y, z));
	PRINT3("object size (%f %f %f)\n", x, y, z);

	return true;
}


bool
ParserIob::ParseBbox(uint32 Len)
{
	ASSERT(Len == 24);

	float			x0, y0, z0;
	ASSERT(ParseVector(x0, y0, z0));
	float			x1, y1, z1;
	ASSERT(ParseVector(x1, y1, z1));

	PRINT6("object bounding box (%f %f %f) to (%f %f %f)\n", x0, y0, z0, x1, y1, z1);

	return true;
}



//---------------------------------
// Stnd-related routines
//---------------------------------

bool
ParserIob::ParseStnd(uint32 Len)
{
	uint32			CurrLen = 0;
	while (CurrLen < Len)
	{
		READ_CHUNK;
		if (ChunkId == OBJ_STID)
		{
			ASSERT2(ParseStid(ChunkLen), "stid");
		}
		else if (ChunkId == OBJ_STDT)
		{
			ASSERT2(ParseStdt(ChunkLen), "stdt");
		}
		else
		{
			SKIP_CHUNK;
		}

		CurrLen += ChunkLen + 2*sizeof(ChunkId);
	}

	PAD_CHUNK;
	return true;
}


bool
ParserIob::ParseStid(uint32 Len)
{
	ASSERT(Len == 20);

	char			Name[MAX_NAMELEN+1];
	ASSERT(ParseName(MAX_NAMELEN, Name));

	uint16			Flags;
	READ_INT16(Flags);
	PrintStidFlags(Flags);
	_PRINT0("\n");

	return true;
}


void
ParserIob::PrintStidFlags(uint16 Flags)
{
	if (Flags == STID_AXIS_DATA)
	{
		_PRINT0("axis data");
	}
	else if (Flags == STID_SHAPE_DATA)
	{
		_PRINT0("shape data");
	}
	else if (Flags == STID_COLOR_DATA)
	{
		_PRINT0("color/properties data");
	}
	else
	{
		_PRINT0("<unknown data>");
	}
}


//---------------------------------
// Stdt-related routines
//---------------------------------

bool
ParserIob::ParseStdt(uint32 Len)
{
	uint16			Tag;
	READ_INT16(Tag);
	uint16			Flags;
	READ_INT16(Flags);

	uint32			LenLeft = Len-4;

	if (Tag == STDT_AXIS_DATA)
	{
		ASSERT2(ParseStdtAxis(LenLeft), "stdt axys");
	}
	else if (Tag == STDT_VERTEX_DATA)
	{
		ASSERT2(ParseStdtVertex(LenLeft), "stdt vertex");
	}
	else if (Tag == STDT_PATH_DATA)
	{
		ASSERT2(ParseStdtPath(LenLeft), "stdt path");
	}
	else if (Tag == STDT_AXIS_SIZE_DATA)
	{
		ASSERT2(ParseStdtAxisSize(LenLeft), "stdt axis");
	}
	else if (Tag == STDT_FACE_COLOR_DATA)
	{
		ASSERT2(ParseStdtFaceColor(LenLeft), "stdt face color");
	}
	else if (Tag == STDT_FACE_FILTER_DATA)
	{
		ASSERT2(ParseStdtFaceFilter(LenLeft), "stdt face filter");
	}
	else if (Tag == STDT_FACE_REFLECT_DATA)
	{
		ASSERT2(ParseStdtReflect(LenLeft), "stdt face reflect");
	}
	else if (Tag == STDT_OBJ_ATTRIB_DATA)
	{
		ASSERT2(ParseStdtObjAttrib(LenLeft), "stdt obj attrib");
	}
	else
	{
		ASSERT(SkipBytes(LenLeft));
	}

	return true;
}


bool
ParserIob::ParseStdtAxis(uint32 Len)
{
	ASSERT(Len == 48);
	float			x, y, z;
	ASSERT(ParseVector(x, y, z));
	PRINT3("position vector (%f %f %f)\n", x, y, z);
	ASSERT(ParseVector(x, y, z));
	PRINT3("alignement vector (%f %f %f)\n", x, y, z);

	return true;
}


bool
ParserIob::ParseStdtVertex(uint32 Len)
{
	uint32			NbVertices = Len / 12;
	ASSERT(NbVertices*12 == Len);

	for (uint32 i = 0; i < NbVertices; i++)
	{
		float			x, y, z;
		ASSERT(ParseVector(x, y, z));
		PRINT3("v (%f %f %f)\n", x, y, z);
	}

	return true;
}


bool
ParserIob::ParseStdtPath(uint32 Len)
{
	uint32			NbPaths = Len / 68;
	ASSERT(NbPaths*68 == Len);

	for (uint32 i = 0; i < NbPaths; i++)
	{
		char		Path[69];
		ASSERT(ReadBytes(68, Path));
		PRINT1("path '%s'\n", Path);
	}

	return true;
}


bool
ParserIob::ParseStdtAxisSize(uint32 Len)
{
	ASSERT(Len == 12);
	float			x, y, z;
	ASSERT(ParseVector(x, y, z));
	PRINT3("obj size %f %f %f\n", x, y, z);

	return true;
}


bool
ParserIob::ParseStdtFaceColor(uint32 Len)
{
	uint32			NbFaces = Len / 3;
	ASSERT(NbFaces*3 == Len);

	for (uint32 i = 0; i < NbFaces; i++)
	{
		uchar		r, g, b;
		ASSERT(ParseColor(r, g, b));
		PRINT3("color %d %d %d\n", r, g, b);
	}

	PAD_CHUNK;
	return true;
}


bool
ParserIob::ParseStdtFaceFilter(uint32 Len)
{
	uint32			NbFaces = Len / 3;
	ASSERT(NbFaces*3 == Len);

	for (uint32 i = 0; i < NbFaces; i++)
	{
		uchar		r, g, b;
		ASSERT(ParseColor(r, g, b));
		PRINT3("face filter %d %d %d\n", r, g, b);
	}

	PAD_CHUNK;
	return true;
}


bool
ParserIob::ParseStdtReflect(uint32 Len)
{
	uint32			NbFaces = Len / 3;
	ASSERT(NbFaces*3 == Len);

	for (uint32 i = 0; i < NbFaces; i++)
	{
		uchar		r, g, b;
		ASSERT(ParseColor(r, g, b));
		PRINT3("face reflect %d %d %d\n", r, g, b);
	}

	PAD_CHUNK;
	return true;
}


bool
ParserIob::ParseStdtObjAttrib(uint32 Len)
{
	ASSERT(Len == 44);
	PRINT0("obj attrib: ");

	uchar			Props[8];
	ASSERT(ReadBytes(8, (char* )Props));
	PrintPrp1(Props);

	uint16			Lamp;
	READ_INT16(Lamp);
	PrintShp2Lamp(Lamp);
	_PRINT0("\n");

	uint16			Flags;
	READ_INT16(Flags);

	float			Fr, Fg, Fb;
	ASSERT(ParseVector(Fr, Fg, Fb));
	PRINT3("intensity= %f %f %f\n", Fr, Fg, Fb);

	float			Fog;
	ASSERT(ParseFract(Fog));
	PRINT1("fog= %f\n", Fog);

	uchar			r, g, b;
	ASSERT(ParseColorPad(r, g, b));
	PRINT3("object color= %d %d %d\n", r, g, b);
	ASSERT(ParseColorPad(r, g, b));
	PRINT3("object reflectance= %d %d %d\n", r, g, b);
	ASSERT(ParseColorPad(r, g, b));
	PRINT3("object transparency= %d %d %d\n", r, g, b);
	ASSERT(ParseColorPad(r, g, b));
	PRINT3("object specularity= %d %d %d\n", r, g, b);

	return true;
}



// --------------------------------------
// --------------------------------------

bool
ParserIob::ParsePnts(uint32 Len)
{
	uint16			NbPoints;
	READ_INT16(NbPoints);
	ASSERT(Len == 2+NbPoints*12);

	for (uint32 i = 0; i < NbPoints; i++)
	{
		float			x, y, z;
		ASSERT(ParseVector(x, y, z));
		PRINT3(" %f %f %f\n", x, y, z);
	}

	return true;
}


bool
ParserIob::ParseEdge(uint32 Len)
{
	uint16			NbEdges;
	READ_INT16(NbEdges);
	ASSERT(Len == 2+NbEdges*4);

	for (uint32 i = 0; i < NbEdges; i++)
	{
		uint16			Start, End;
		READ_INT16(Start);
		READ_INT16(End);
		PRINT2("edge from %d to %d\n", Start, End);
	}

	return true;
}


bool
ParserIob::ParseFace(uint32 Len)
{
	uint16			NbFaces;
	READ_INT16(NbFaces);
	ASSERT(Len == 2+NbFaces*6);

	for (uint32 i = 0; i < NbFaces; i++)
	{
		uint16			i0, i1, i2;
		READ_INT16(i0);
		READ_INT16(i1);
		READ_INT16(i2);
		PRINT3("face %d %d %d\n", i0, i1, i2);
	}

	return true;
}


//-----------------------------------
// Pth2-related routines
//-----------------------------------

bool
ParserIob::ParsePth2(uint32 Len)
{
	uint16			NbAxis;
	READ_INT16(NbAxis);
	ASSERT(Len == 2+NbAxis*68);

	for (uint32 i = 0; i < NbAxis; i++)
	{
		ASSERT2(ParsePthd(68), "pthd");
	}

	return true;
}


bool
ParserIob::ParsePthd(uint32 Len)
{
	ASSERT(Len == 68);
	ASSERT2(ParseTform(60), "pthd tform");
	uint16			inFrom;
	READ_INT16(inFrom);
	uint16			OutTo;
	READ_INT16(OutTo);
	uint16			Flags;
	READ_INT16(Flags);
	uint16			Pad;
	READ_INT16(Pad);
	PRINT3("inFrom=%d, OutTo=%d, Flags=%x\n", inFrom, OutTo, Flags);

	return true;
}


//-----------------------------------------
//-----------------------------------------

bool
ParserIob::ParseColr(uint32 Len)
{
	ASSERT(Len == 4);
	uchar			r, g, b;
	ASSERT(ParseColorPad(r, g, b));
	PRINT3("colr (%d %d %d)\n", r, g, b);

	return true;
}


bool
ParserIob::ParseRefl(uint32 Len)
{
	ASSERT(Len == 4);
	uchar			r, g, b;
	ASSERT(ParseColorPad(r, g, b));
	PRINT3("refl (%d %d %d)\n", r, g, b);

	return true;
}


bool
ParserIob::ParseTran(uint32 Len)
{
	ASSERT(Len == 4);
	uchar			r, g, b;
	ASSERT(ParseColorPad(r, g, b));
	PRINT3("tran (%d %d %d)\n", r, g, b);

	return true;
}


bool
ParserIob::ParseSpc1(uint32 Len)
{
	ASSERT(Len == 4);
	uchar			r, g, b;
	ASSERT(ParseColorPad(r, g, b));
	PRINT3("spc1 (%d %d %d)\n", r, g, b);

	return true;
}


//------------------------------------
//------------------------------------


bool
ParserIob::ParseClst(uint32 Len)
{
	uint32			NbFaces = Len / 3;
	ASSERT(NbFaces*3 == Len);

	for (uint32 i = 0; i < NbFaces; i++)
	{
		uchar		r, g, b;
		ASSERT(ParseColor(r, g, b));
		PRINT3("clst %d %d %d\n", r, g, b);
	}

	PAD_CHUNK;
	return true;
}


bool
ParserIob::ParseRlst(uint32 Len)
{
	uint32			NbFaces = Len / 3;
	ASSERT(NbFaces*3 == Len);

	for (uint32 i = 0; i < NbFaces; i++)
	{
		uchar		r, g, b;
		ASSERT(ParseColor(r, g, b));
		PRINT3("rlst %d %d %d\n", r, g, b);
	}

	PAD_CHUNK;
	return true;
}


bool
ParserIob::ParseTlst(uint32 Len)
{
	uint32			NbFaces = Len / 3;
	ASSERT(NbFaces*3 == Len);

	for (uint32 i = 0; i < NbFaces; i++)
	{
		uchar		r, g, b;
		ASSERT(ParseColor(r, g, b));
		PRINT3("tlst %d %d %d\n", r, g, b);
	}

	PAD_CHUNK;
	return true;
}


//---------------------------------
//---------------------------------
bool
ParserIob::ParseTxt3(uint32 Len)
{
	PRINT0("texture: ");

	uint16				Flags;
	READ_INT16(Flags);
	if (Flags & TXT3_APPLY_TO_CHILDREN)
	{
		_PRINT0("apply to children\n");
	}

	ASSERT2(ParseTform(60), "txt3 pthd");

	// Texture parameters
	for (int32 i = 0; i < 16; i++)
	{
		float			Param;
		ASSERT(ParseFract(Param));
	}

	// Texture flags
	uchar			Pflags[16];
	ASSERT(ReadBytes(16, (char* )Pflags));

	// Texture subgroups
	uchar			SubGroup[18];
	ASSERT(ReadBytes(18, (char* )SubGroup));

	// Texture file name
	uchar			TxtLen;
	char*			TxtFilename;
	READ_INT8(TxtLen);

	int32			TestLen = (Len & 0x1)?1:0;
	TestLen += 2+60+16*4+16+18+18+1+TxtLen;
	ASSERT(TestLen == Len);

	ASSERT(LexString(int32(TxtLen), TxtFilename));
	PRINT1("texture '%s'\n", TxtFilename);
	DeleteLexeme(TxtFilename);

	PAD_CHUNK;
	return true;
}


//--------------------------------------
// Brs-related routines
//--------------------------------------

bool
ParserIob::ParseBrs4(uint32 Len)
{
	uint16			Flags;
	READ_INT16(Flags);
	PrintBrs4Flags(Flags);

	uint16			Wflags;
	READ_INT16(Wflags);
	PrintBrs4Wflags(Flags);

	ASSERT2(ParseTform(60), "brs4 tform");

	uint16			FullScale;
	READ_INT16(FullScale);
	uint16			MaxSeq;
	READ_INT16(MaxSeq);
	PRINT2("FullScale=%d, MaxSeq=%d\n", FullScale, MaxSeq);

	uchar			SubGroup[18];
	ASSERT(ReadBytes(18, (char* )SubGroup));

	uchar			LockState[18];
	ASSERT(ReadBytes(18, (char* )LockState));

	// Texture file name
	uchar			BrushLen;
	char*			BrushFilename;
	READ_INT8(BrushLen);

	int32			TestLen = (Len & 0x1)?1:0;
	TestLen += 4+60+4+18*2+1+BrushLen;
	ASSERT(TestLen == Len);

	ASSERT(LexString(BrushLen, BrushFilename));
	PRINT1("brush '%s'\n", BrushFilename);
	DeleteLexeme(BrushFilename);

	PAD_CHUNK;
	return true;
}


void
ParserIob::PrintBrs4Flags(uint16 Flags)
{
	if (Flags & BRS4_FLAG_COLOR)
	{
		_PRINT0("color, ");
	}
	if (Flags & BRS4_FLAG_REFLECTIVITY)
	{
		_PRINT0("reflectivity, ");
	}
	if (Flags & BRS4_FLAG_FILTER)
	{
		_PRINT0("filter, ");
	}
	if (Flags & BRS4_FLAG_ALTITUDE)
	{
		_PRINT0("altitude, ");
	}
	if (Flags & BRS4_FLAG_REFLECTION)
	{
		_PRINT0("reflection, ");
	}
}


void
ParserIob::PrintBrs4Wflags(uint16 Flags)
{
	if (Flags & BRS4_WFLAG_WRAP_X)
	{
		_PRINT0("wrap x, ");
	}
	if (Flags & BRS4_WFLAG_WRAP_Z)
	{
		_PRINT0("wrap z, ");
	}
	if (Flags & BRS4_WFLAG_WRAP_APPLY_TO_CHILDREN)
	{
		_PRINT0("apply to children, ");
	}
	if (Flags & BRS4_WFLAG_WRAP_MIRROR)
	{
		_PRINT0("mirror, ");
	}
	if (Flags & BRS4_WFLAG_WRAP_INVERSE_VIDEO)
	{
		_PRINT0("inverse video, ");
	}
	if (Flags & BRS4_WFLAG_WRAP_USE_GENLOCK)
	{
		_PRINT0("use genlock, ");
	}
}


bool
ParserIob::ParseFogl(uint32 Len)
{
	ASSERT(Len == 4);
	float			Fog;
	ASSERT(ParseFract(Fog));
	PRINT1("fog %f\n", Fog);

	return true;
}



bool
ParserIob::ParsePrp1(uint32 Len)
{
	ASSERT(Len == 8);
	uchar			Props[8];
	ASSERT(ReadBytes(8, (char* )Props));
	float			RefractIndex = float(Props[4])*100.0+1.0;
	PRINT4("dither=%d, hardness=%d, roughness=%d, shinyness=%d\n",
			Props[0], Props[1], Props[2], Props[3]);
	PRINT4("RefractIndex=%d, QuickDraw=%s, Phong=%s, Genlock=%s",
			RefractIndex, QuickDraw[Props[5]], Props[6]?"on":"off",
			Props[7]?"on":"off");

	return true;
}



bool
ParserIob::ParseInt1(uint32 Len)
{
	ASSERT(Len == 12);
	float			r, g, b;
	ASSERT(ParseVector(r, g, b));
	PRINT3("light intensity (%f %f %f)\n", r, g, b);

	return true;
}


bool
ParserIob::ParseAnid(uint32 Len)
{
	ASSERT(Len == 64);
	uint32			CellNumber;
	READ_INT32(CellNumber);
	PRINT1("anid: cell number=%d\n", CellNumber);
	ASSERT2(ParseTform(60), "anid tform");

	return true;
}


//------------------------------------------
// For2-related routines
//------------------------------------------

bool
ParserIob::ParseFor2(uint32 Len)
{
	uint16			NumC;
	READ_INT16(NumC);

	uint16			NumF;
	READ_INT16(NumF);
	PRINT2("NumC=%d, NumF=%d\n", NumC, NumF);

	uint16			Mult;
	uint16			Flags;
	READ_INT16(Flags);
	PrintFor2Flags(Flags, Mult);

	uint16			NbKeys;
	READ_INT16(NbKeys);

	uint32			Pc = NbKeys*NumC + Mult*NumF;
	ASSERT(56 + 12*Pc + 2*NbKeys == Len);

	PRINT0("object rotation: ");
	ASSERT2(ParseMatrix(), "for2 matrix");

	float			tx, ty, tz;
	ASSERT(ParseVector(tx, ty, tz));
	PRINT3("object translation: %f %f %f", tx, ty, tz);

	for (int32 i = 0; i < Pc; i++)
	{
		float			x, y, z;
		ASSERT(ParseVector(x, y, z));
		PRINT3("form point: %f %f %f", x, y, z);
	}

	PRINT0("sections: ");
	for (int32 i = 0; i < NbKeys; i++)
	{
		uint16			Section;
		READ_INT16(Section);
		_PRINT1("%d ", Section);
	}
	_PRINT0("\n");

	return true;
}


void
ParserIob::PrintFor2Flags(uint16 Flags, uint16& Out)
{
	if (Flags & FOR2_XY_CROSS_SECTION)
	{
		PRINT0("x/y cross section, ");
	}
	else
	{
		PRINT0("y/z cross section, ");
	}

	if (!((Flags & FOR2_ONE_FORMER_VIEW) || (Flags & FOR2_ONE_SPACER_VIEW)))
	{
		_PRINT0("2 former view");
		Out = 4;
	}
	else if (Flags & FOR2_ONE_FORMER_VIEW)
	{
		_PRINT0("1 former view, ");
		Out = 2;
	}
	else if (Flags & FOR2_ONE_SPACER_VIEW)
	{
		_PRINT0("1 spacer view, ");
		Out = 1;
	}
	_PRINT0("\n");
}



//------------------------------------------
// Part-related routines
//------------------------------------------

bool
ParserIob::ParsePart(uint32 Len)
{
	ASSERT(Len == 6);
	uint16			Type;
	READ_INT16(Type);
	PrintPartType(Type);

	float			Size;
	ASSERT(ParseFract(Size));
	_PRINT1("size=%f\n", Size);

	return true;
}


void
ParserIob::PrintPartType(uint16 Type)
{
	_PRINT1("shape=%s, ", PartShape[Type & 0x000f]);
	_PRINT1("centering=%s, ", PartCenter[(Type >> 4) & 0x000f]);
	_PRINT1("size=%s, ", PartSize[(Type >> 8) & 0x000f]);
	_PRINT1("alignment=%s, ", PartAlign[(Type >> 12) & 0x000f]);
}



bool
ParserIob::ParsePtfn(uint32 Len)
{
	uchar			PartLen;
	char*			PartFile;
	READ_INT8(PartLen);

	int32			TestLen = (Len & 0x1)?1:0;
	TestLen += 1+PartLen;
	ASSERT(TestLen == Len);

	ASSERT(LexString(PartLen, PartFile));
	DeleteLexeme(PartFile);

	PAD_CHUNK;
	return true;
}



bool
ParserIob::ParseFgr2(uint32 Len)
{
	uint16			NbFaces;
	READ_INT16(NbFaces);

	char			SubGroupName[18];
	ASSERT(ReadBytes(18, SubGroupName));
	PRINT1("name '%s'\n", SubGroupName);

	for (int32 i = 0; i < NbFaces; i++)
	{
		uint16		Index;
		READ_INT16(Index);
		PRINT1("%d ", Index);
	}
	_PRINT0("\n");

	uint16			Type;
	READ_INT16(Type);
	PrintPartType(Type);
	float			Size;
	ASSERT(ParseFract(Size));

	uchar			PartLen;
	char*			PartFile;
	READ_INT8(PartLen);

	int32			TestLen = (Len & 0x1)?1:0;
	TestLen += 20 + NbFaces*2+7+PartLen;
	ASSERT(TestLen == Len);

	ASSERT(LexString(PartLen, PartFile));
	PRINT1("part filename '%s'\n", PartFile);
	DeleteLexeme(PartFile);

	PAD_CHUNK;
	return true;
}



bool
ParserIob::ParseBbsg(uint32 Len)
{
	ASSERT(Len == 18);
	char			Bbsg[18];
	ASSERT(ReadBytes(18, Bbsg));

	return true;
}



bool
ParserIob::ParseSbsg(uint32 Len)
{
	ASSERT(Len == 18);
	char			Bbsg[18];
	ASSERT(ReadBytes(18, Bbsg));

	return true;
}



bool
ParserIob::ParseEflg(uint32 Len)
{
	uint16			NbEdges;
	READ_INT16(NbEdges);
	PRINT1("eflg %d: ", NbEdges);

	for (int32 i = 0; i < NbEdges; i++)
	{
		uchar			EdgeFlag;
		READ_INT8(EdgeFlag);
		if (EdgeFlag & EFLG_QUICK_EDGE)
		{
			_PRINT0("quick");
		}
		else if (EdgeFlag & EFLG_SHARP_EDGE)
		{
			_PRINT0("sharp");
		}
		_PRINT0(", ");
	}
	_PRINT0("\n");

	return true;
}

*/


